/* FUNCTION: divers.fonction_controle_doublons(text[], integer, boolean)
Fonction permettant  de contrôler les doublons au sein d'un jeu de données ou bien entre un jeu de données et tous les autres jeux de données

Modifications apportées entre le v2.0 et la v3.0 (date 30/04/2020) :
- modification du comportement de la fonction, qui regroupe toutes les observations dont les observateurs, le denombrement et le détail des individus sont identiques lors du test de doublon interne
- ajout du nombre d'observations en doublons (mais il y a un bug dans le calcul)

Modifications apportées entre le v1.3 et la v2.0 :
- ajout du vidage de table divers.controle_doublons pour ne pas avoir les résultats d'un test précédent
- ajout d'un paramètre permettant de définit une distance maximale entre deux obs pour qu'elles soient considérées comme des doublons potentiels
- modification du tri des résultats envoyés : celui-ci se fait par id pour pouvoir mieux gérer les triplons, étant donné que de toutes façons on définit une distance max en paramètre dans la fonction.

*/
-----------------------------------------------
-- 1. Création de vues et tables préalables
-----------------------------------------------

-- 1.1 On crée la vue occtax.v_descriptif_sujet_decodee
-------------------------------------------------------

DROP VIEW IF EXISTS occtax.v_descriptif_sujet_decodee;
CREATE OR REPLACE VIEW occtax.v_descriptif_sujet_decodee AS
SELECT o.cle_obs,
	nobsmet.valeur AS obs_methode,
	nocetatbio.valeur AS occ_etat_biologique,
	nocnat.valeur AS occ_naturalite,
	nsex.valeur AS occ_sexe,
	nstade.valeur AS occ_stade_de_vie,
	nbiogeo.valeur AS occ_statut_biogeographique,
	nstatbio.valeur AS occ_statut_biologique,
	npreuv.valeur AS preuve_existante,
	ds.preuve_numerique,
	ds.preuve_non_numerique,
	ds.obs_contexte,
	ds.obs_description,
	ds.occ_methode_determination,
	ds.occ_denombrement_min,
	ds.occ_denombrement_max,
	nobjden.valeur AS occ_objet_denombrement,
	ntypden.valeur AS occ_type_denombrement
FROM occtax.observation o
JOIN LATERAL jsonb_to_recordset(o.descriptif_sujet) ds(obs_methode text, occ_etat_biologique text, occ_naturalite text, occ_sexe text, occ_stade_de_vie text, occ_statut_biogeographique text, occ_statut_biologique text, preuve_existante text, preuve_numerique text, preuve_non_numerique text, obs_contexte text, obs_description text, occ_methode_determination text, occ_denombrement_min text, occ_denombrement_max text, occ_objet_denombrement text, occ_type_denombrement text) ON true
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_etat_biologique') nocetatbio ON ds.occ_etat_biologique = nocetatbio.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_naturalite') nocnat ON ds.occ_naturalite = nocnat.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_sexe') nsex ON ds.occ_sexe = nsex.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'obs_methode') nobsmet ON ds.obs_methode = nobsmet.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_stade_de_vie') nstade ON ds.occ_stade_de_vie = nstade.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_statut_biogeographique') nbiogeo ON ds.occ_statut_biogeographique = nbiogeo.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_statut_biologique') nstatbio ON ds.occ_statut_biologique = nstatbio.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'preuve_existante') npreuv ON ds.preuve_existante = npreuv.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_type_denombrement') ntypden ON ds.occ_type_denombrement = ntypden.code
LEFT JOIN (SELECT * FROM occtax.nomenclature WHERE champ ilike 'occ_objet_denombrement') nobjden ON ds.occ_objet_denombrement = nobjden.code
;

COMMENT ON VIEW v_descriptif_sujet_decodee IS 'Vue permettant d''éclater en tableau et de décoder les valeurs de descriptif_sujet';
ALTER TABLE occtax.v_descriptif_sujet_decodee OWNER TO postgres;
GRANT ALL ON TABLE occtax.v_descriptif_sujet_decodee TO postgres;
GRANT ALL ON TABLE occtax.v_descriptif_sujet_decodee TO lizmap;

DROP TABLE IF EXISTS divers.controle_doublons ;

-- 1.2 On crée la table divers.controle_doublons
-------------------------------------------------------
DROP TABLE IF EXISTS divers.controle_doublons ;
CREATE TABLE divers.controle_doublons (
	jdd_code TEXT,
	jdd_id TEXT,
	organisme_gestionnaire_donnees TEXT,
	cle_obs TEXT,
	identifiant_origine TEXT,
	identifiant_permanent TEXT,
	nb_obs INTEGER,
	distance_geom_m NUMERIC,
	group2_inpn TEXT,
	nom_valide TEXT,
	nom_vern TEXT,
	date_debut DATE,
	observateurs TEXT,
	denombrement_total TEXT,
	detail_individus TEXT,
	validite_niveau TEXT,
	validite_date_validation TEXT,
	dernier_import TEXT,
	dee_date_derniere_modification TEXT
) ;

COMMENT ON TABLE divers.controle_doublons IS 'Table stockant le résultat de la fonction de recherche de doublons divers.fonction_controle_doublons(text[], integer, boolean)';

ALTER TABLE divers.controle_doublons OWNER TO postgres;
GRANT ALL ON TABLE divers.controle_doublons TO postgres;
GRANT ALL ON TABLE divers.controle_doublons TO lizmap;


-----------------------------------------------
-- 2. Création de la fonction à proprement parler
-----------------------------------------------

-- DROP FUNCTION divers.fonction_controle_doublons(text[], integer, boolean);
CREATE OR REPLACE FUNCTION divers.fonction_controle_doublons(
	jdd_id text[],
	rayon_en_m INTEGER,
	p_controle_interne BOOLEAN)
    RETURNS void
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE 
AS $BODY$


DECLARE sql_requete_principale TEXT ;
DECLARE sql_clause_where TEXT ;
DECLARE sql_group_by TEXT ;
DECLARE sql_insert TEXT ;
DECLARE jdd_id_compares TEXT[] ;

BEGIN

	RAISE NOTICE 'Je suis en train de travailler, ça peut prendre quelques minutes ! L''occasion de pratiquer la respiration ventrale...' ;
	
-- 2.1 On vide la table divers.controle_doublons 
-------------------------------------------------------
	DELETE FROM divers.controle_doublons ;

-- 2.2 On établit la liste des jdd_id auxquels comparer le(s) jdd testé(s)
-------------------------------------------------------
	
	IF p_controle_interne IS TRUE 
		THEN SELECT INTO jdd_id_compares array_agg(jdd.jdd_id) FROM occtax.jdd WHERE jdd.jdd_id = ANY ($1) ;
		ELSE SELECT INTO jdd_id_compares array_agg(jdd.jdd_id) FROM occtax.jdd WHERE NOT(jdd.jdd_id = ANY ($1)) ;
	END IF ;
	
	
-- 2.3 On lance la fonction
--------------------------

	sql_requete_principale := '

	-- Table complète des obs, enrichie par jointures
	WITH v_obs AS (
			SELECT 	o.cle_obs,
				o.jdd_code,
				o.organisme_gestionnaire_donnees,			
				o.jdd_id,
				o.identifiant_origine,
				o.identifiant_permanent,
				COALESCE(o.geom, m02.geom, m10.geom) AS geom, -- on prend la géométrie de l''objet et à défaut celle de la maille la plus précise
				o.date_debut,
				o.cd_ref,
				o.denombrement_min,
				string_agg(DISTINCT
					CONCAT_WS ('' '',
							ds.occ_denombrement_min, lower(ds.occ_objet_denombrement) || ''(s)'',
							CASE WHEN ds.occ_etat_biologique IN (''Trouvé mort'') THEN lower(ds.occ_etat_biologique) ELSE NULL END,
							CASE WHEN ds.occ_sexe NOT IN (''Inconnu'', ''Non renseigné'', ''NSP'', ''Indéterminé'') THEN ''de sexe '' || lower(ds.occ_sexe) ELSE NULL END,
							CASE WHEN ds.occ_stade_de_vie NOT IN (''Inconnu'', ''Non renseigné'', ''NSP'', ''Indéterminé'') THEN ''au stade '' || lower(ds.occ_stade_de_vie) ELSE NULL END,
							CASE WHEN ds.occ_statut_biologique NOT IN (''Inconnu'', ''Non renseigné'', ''NSP'', ''Indéterminé'') THEN ''de statut '' || lower(ds.occ_statut_biologique) ELSE NULL END
						   )
					, '' ; ''
					) AS detail_individus,
				string_agg(DISTINCT p.identite, '' - '' ORDER BY p.identite) AS observateurs,
				o.validite_niveau,
				o.validite_date_validation,
				o.dee_date_derniere_modification::DATE,
				max(i.date_import) AS dernier_import
		FROM occtax.observation o
		LEFT JOIN occtax.observation_personne op USING(cle_obs) 
		LEFT JOIN occtax.personne p USING(id_personne) 
		-- ajout de jointures pour récupérer les mailles
		LEFT JOIN occtax.localisation_maille_02 lm02 ON lm02.cle_obs = o.cle_obs
		LEFT JOIN sig.maille_02 m02 ON lm02.code_maille = m02.code_maille
		LEFT JOIN occtax.localisation_maille_10 lm10 ON lm10.cle_obs=o.cle_obs
		LEFT JOIN sig.maille_10 m10 ON lm10.code_maille = m10.code_maille
		LEFT JOIN occtax.jdd_import i ON o.jdd_id=i.jdd_id
		LEFT JOIN occtax.v_descriptif_sujet_decodee ds ON ds.cle_obs = o.cle_obs
		WHERE (op.role_personne=''Obs'' OR op.role_personne IS NULL)
		GROUP BY o.cle_obs, o.jdd_code, o.jdd_id, o.identifiant_origine, COALESCE(o.geom, m02.geom, m10.geom), o.date_debut, o.cd_ref, o.organisme_gestionnaire_donnees, o.denombrement_min
			)
			
	-- Table(s) qu''on souhaite tester
	,obs_source AS (
		SELECT * FROM v_obs WHERE v_obs.jdd_id = ANY ($1)
		)
		
	-- Tables auxquelles on souhaite la comparer
	,obs_cible AS (
		SELECT * FROM v_obs WHERE v_obs.jdd_id = ANY ($2)
		)
	
	,liste_doublons_brute AS (
		SELECT 
			CASE WHEN s.jdd_code = c.jdd_code 
				THEN s.jdd_code
				ELSE CONCAT(s.jdd_code, '' | '', c.jdd_code) 
			END AS jdd_code, 
			
			CASE WHEN s.jdd_id = c.jdd_id 
				THEN s.jdd_id
				ELSE CONCAT(s.jdd_id, '' | '', c.jdd_id) 
			END AS jdd_id, 

			CASE 
				WHEN TRIM(lower(s.organisme_gestionnaire_donnees)) = TRIM(LOWER(c.organisme_gestionnaire_donnees))
				THEN CONCAT(''Même organisme'', '' ('', s.organisme_gestionnaire_donnees, '')'')
				ELSE CONCAT(s.organisme_gestionnaire_donnees, '' | '', c.organisme_gestionnaire_donnees)
			END AS organisme_gestionnaire_donnees,	
			
			s.cle_obs AS cle_obs_source, 
			c.cle_obs AS cle_obs_cible,
			s.identifiant_origine AS identifiant_origine_source,
			c.identifiant_origine AS identifiant_origine_cible,
			s.identifiant_permanent AS identifiant_permanent_source,
			c.identifiant_permanent AS identifiant_permanent_cible,
			st_distance(s.geom, c.geom)::NUMERIC(10,1) AS distance_geom_m,
			t.group2_inpn,
			t.nom_valide, 
			t.nom_vern,
			c.date_debut,
				
			CASE WHEN s.observateurs = c.observateurs 
				THEN CONCAT(''Mêmes observateurs'', '' ('', s.observateurs, '')'')
				ELSE CONCAT(s.observateurs, '' | '', c.observateurs)
			END AS observateurs,
			
			CASE WHEN s.denombrement_min = c.denombrement_min 
				THEN CONCAT(''Même effectif'', '' ('', s.denombrement_min, '')'')
				ELSE CONCAT(s.denombrement_min, '' | '', c.denombrement_min)
			END AS denombrement_total,
			
			CASE WHEN s.detail_individus = c.detail_individus 
				THEN CONCAT(''Mêmes individus'', '' ('', s.detail_individus, '')'')
				ELSE CONCAT(s.detail_individus, '' | '', c.detail_individus)
			END AS detail_individus,

			CASE WHEN s.validite_niveau = c.validite_niveau 
				THEN CONCAT(''Même niveau'', '' ('', s.validite_niveau, '')'')
				ELSE CONCAT(s.validite_niveau, '' | '', c.validite_niveau)
			END AS validite_niveau,
			
			CASE WHEN s.validite_date_validation = c.validite_date_validation 
				THEN s.validite_date_validation::TEXT
				ELSE CONCAT(s.validite_date_validation, '' | '', c.validite_date_validation)
			END AS validite_date_validation, 
			
			CASE WHEN s.dernier_import = c.dernier_import 
				THEN s.dernier_import::TEXT
				ELSE CONCAT(s.dernier_import, '' | '', c.dernier_import)
			END AS dernier_import, 

			CASE WHEN s.dee_date_derniere_modification = c.dee_date_derniere_modification 
				THEN s.dee_date_derniere_modification::TEXT
				ELSE CONCAT(s.dee_date_derniere_modification, '' | '', c.dee_date_derniere_modification)
			END AS dee_date_derniere_modification
			
		FROM obs_source s
		INNER JOIN obs_cible c 
			ON s.cd_ref = c.cd_ref 
			AND s.date_debut = c.date_debut 
			AND st_DWithin(s.geom, c.geom, $3) -- les géométries doivent être distantes de moins de la valeur indiquée en paramètre
		LEFT JOIN taxon.taxref_valide t ON s.cd_ref = t.cd_ref
		WHERE s.cle_obs <> c.cle_obs -- pour les cas où on teste intra-jeu
	';
	
	IF p_controle_interne IS TRUE 
		THEN sql_clause_where = 
				'AND s.cle_obs < c.cle_obs
				'; -- permet de ne pas lister deux fois une même paire d'observations
		ELSE sql_clause_where = 
				'AND TRUE
				' ;
	END IF ;
	
	sql_group_by = '
	GROUP BY 	s.jdd_code, c.jdd_code, s.jdd_id, c.jdd_id, s.cle_obs, c.cle_obs, c.identifiant_origine, s.identifiant_origine, s.identifiant_permanent, c.identifiant_permanent, c.organisme_gestionnaire_donnees, s.organisme_gestionnaire_donnees, 
	c.date_debut, t.nom_valide, t.group2_inpn, t.nom_vern, c.observateurs, s.observateurs, c.dernier_import, s.dernier_import, c.dee_date_derniere_modification, s.dee_date_derniere_modification, c.validite_niveau, s.validite_niveau, c.validite_date_validation, s.validite_date_validation, s.denombrement_min, c.denombrement_min, s.geom, c.geom, s.detail_individus, c.detail_individus
	ORDER BY s.identifiant_origine, c.identifiant_origine, st_distance(s.geom, c.geom)::NUMERIC(10,1) --  Finalement, on trie par id pour pouvoir mieux gérer les triplons, étant donné que de toutes façons on définit une distance max en paramètre dans la fonction.
	' ;
	
	-- il est ensuite nécessaire de distinguer les cas : pour les doublons internes strictement identiques on regroupe les obs, pour les doublons externes ou les doublons internes non strictement identiques on garde des paires d'observation
	IF p_controle_interne IS TRUE 	
		THEN sql_insert = 
			'
				)		
			--Insertion
			INSERT INTO divers.controle_doublons(jdd_code, jdd_id, organisme_gestionnaire_donnees, cle_obs, identifiant_origine, identifiant_permanent, nb_obs, distance_geom_m, group2_inpn, nom_valide, nom_vern, date_debut, observateurs, denombrement_total, detail_individus, validite_niveau, validite_date_validation, dernier_import, dee_date_derniere_modification)
			
			-- On prend déjà les obs a priori complétement identiques
			SELECT jdd_code, jdd_id, organisme_gestionnaire_donnees, 
			CONCAT(
				string_agg(DISTINCT cle_obs_source::TEXT, '', '' ORDER BY cle_obs_source::TEXT),
				'', '',
				string_agg(DISTINCT cle_obs_cible::TEXT, '','' ORDER BY cle_obs_cible::TEXT)
				)AS cle_obs, 
			
			CONCAT(
				string_agg(DISTINCT identifiant_origine_source, '', '' ORDER BY identifiant_origine_source),
				'', '',
				string_agg(DISTINCT identifiant_origine_cible, '','' ORDER BY identifiant_origine_cible)
				)AS identifiant_origine, 
			
			CONCAT(
				string_agg(DISTINCT identifiant_permanent_source, '', '' ORDER BY identifiant_permanent_source),
				'', '',
				string_agg(DISTINCT identifiant_permanent_cible, '','' ORDER BY identifiant_permanent_cible)
				)AS identifiant_permanent, 
			COUNT(DISTINCT cle_obs_cible) + COUNT(DISTINCT cle_obs_source) AS nb_obs,
			distance_geom_m, group2_inpn, nom_valide, nom_vern, date_debut, observateurs, denombrement_total, detail_individus, validite_niveau, validite_date_validation, dernier_import, dee_date_derniere_modification
			
			FROM liste_doublons_brute
			WHERE observateurs ILIKE ''Même%%'' AND denombrement_total ILIKE ''Même%%'' AND detail_individus ILIKE ''Même%%''			
			
			GROUP BY jdd_code, jdd_id, organisme_gestionnaire_donnees, distance_geom_m, group2_inpn, nom_valide, nom_vern, date_debut, observateurs, denombrement_total, detail_individus, validite_niveau, validite_date_validation, dernier_import, dee_date_derniere_modification
			
			-- Puis on complète par les obs qui présentent certaines différences sur les observateurs, le denombrement ou le détail des individus, qu''on ne regroupe pas
			UNION
			SELECT jdd_code, jdd_id, organisme_gestionnaire_donnees, 
			CONCAT(cle_obs_source, '' | '', cle_obs_cible) AS cle_obs, 
			CONCAT(identifiant_origine_source, '' | '', identifiant_origine_cible) AS identifiant_origine, 
			CONCAT(identifiant_permanent_source, '' | '', identifiant_permanent_cible) AS identifiant_permanent, 
			2 AS nb_obs,			
			distance_geom_m, group2_inpn, nom_valide, nom_vern, date_debut, observateurs, denombrement_total, detail_individus, validite_niveau, validite_date_validation, dernier_import, dee_date_derniere_modification
			
			FROM liste_doublons_brute			
			WHERE NOT(observateurs ILIKE ''Même%%'' AND denombrement_total ILIKE ''Même%%'' AND detail_individus ILIKE ''Même%%'')
						
			ORDER BY identifiant_origine
			';
		ELSE sql_insert = 	
			'
			)		
			--Insertion
			INSERT INTO divers.controle_doublons(jdd_code, jdd_id, organisme_gestionnaire_donnees, cle_obs, identifiant_origine, identifiant_permanent, nb_obs, distance_geom_m, group2_inpn, nom_valide, nom_vern, date_debut, observateurs, denombrement_total, detail_individus, validite_niveau, validite_date_validation, dernier_import, dee_date_derniere_modification)
			
			SELECT jdd_code, jdd_id, organisme_gestionnaire_donnees, 
			CONCAT(cle_obs_source, '' | '', cle_obs_cible) AS cle_obs, 
			CONCAT(identifiant_origine_source, '' | '', identifiant_origine_cible) AS identifiant_origine, 
			CONCAT(identifiant_permanent_source, '' | '', identifiant_permanent_cible) AS identifiant_permanent,
			2 AS nb_obs,			
			distance_geom_m, group2_inpn, nom_valide, nom_vern, date_debut, observateurs, denombrement_total, detail_individus, validite_niveau, validite_date_validation, dernier_import, dee_date_derniere_modification
			
			FROM liste_doublons_brute
						
			ORDER BY identifiant_origine_source, identifiant_origine_cible
			';
	END IF;
	
    EXECUTE format(sql_requete_principale) || format(sql_clause_where) || format(sql_group_by) || format(sql_insert)
	USING jdd_id, jdd_id_compares, rayon_en_m ;
	
	
	RAISE NOTICE 'Voilà j''ai fini ! N''oublie pas de lancer la requête SELECT * FROM divers.controle_doublons pour voir le résultat de mon analyse' ;
	
END ;						 

$BODY$;
COMMENT ON FUNCTION divers.fonction_controle_doublons(text[], integer, boolean) IS 'Fonction permettant de lister dans la table divers.controle_doublons les doublons potentiels associés à un ou plusieurs jeux de données. Le premier argument de la fonction est un tableau contenant le(s) jdd_id de()s jeu(x) de données qu''on souhaite tester. LE deuxième argument correspond à la distance maximale en mètres entre observations pour pouvoir être considféré comme un doublon potentiel. Le troisième argument indique si on souhaite un contrôle interne au sein de(s) jdd testé(s), ou bien externe (le ou les jdd testés par rapport à l''ensemble des autres jdd). Par défaut, les obs éloignées de plus de 10 km ne sont pas prises en compte.' ;

ALTER FUNCTION divers.fonction_controle_doublons(text[], integer, boolean)
    OWNER TO postgres;

GRANT EXECUTE ON FUNCTION divers.fonction_controle_doublons(text[], integer, boolean) TO postgres;
GRANT EXECUTE ON FUNCTION divers.fonction_controle_doublons(text[], integer, boolean) TO lizmap;
GRANT EXECUTE ON FUNCTION divers.fonction_controle_doublons(text[], integer, boolean) TO PUBLIC;

-- Test avec 2 jeu de données
SELECT divers.fonction_controle_doublons(ARRAY['27915'], 1, TRUE) ;
SELECT * FROM divers.controle_doublons ;
